@@ -1,13 +1,15 @@
 arch_get_kernel_flavour () {
 	CPU=`grep '^cpu[[:space:]]*:' "$CPUINFO" | head -n1 | cut -d: -f2 | sed 's/^ *//; s/[, ].*//' | tr A-Z a-z`
 	case "$CPU" in
-	    power3|power4|power4+|ppc970*|power5|power5+|power6|power6x|i-star|s-star|pa6t|rs64-*)
+	    power3|power4|power4+|ppc970*|power5|power5+|power6|power6x)
+		family=powerpc64 ;;
+	    cell|i-star|s-star|pa6t|rs64-*)
 		family=powerpc64 ;;
     	    *)
 		family=powerpc ;;
 	esac
 	case "$SUBARCH" in
-	    powermac*|chrp*|pasemi)
+	    powermac*|chrp*|pasemi|ps3)
 		echo "$family" ;;
 	    prep)
 		echo prep ;;
