@@ -16,7 +16,11 @@ DISTRIBUTION=
 # used by kernel installation code
 KERNEL=
 KERNEL_LIST=/tmp/available_kernels.txt
-KERNEL_MAJOR="$(uname -r | cut -d . -f 1,2)"
+case `udpkg --print-os` in
+	linux)		KERNEL_MAJOR="$(uname -r | cut -d . -f 1,2)" ;;
+	kfreebsd)	KERNEL_MAJOR="$(uname -r | cut -d . -f 1)" ;;
+	hurd)		KERNEL_MAJOR="$(uname -v | cut -d ' ' -f 2 | cut -d . -f 1)" ;;
+esac
 KERNEL_VERSION="$(uname -r | cut -d - -f 1)"
 KERNEL_ABI="$(uname -r | cut -d - -f 1,2)"
 KERNEL_FLAVOUR=$(uname -r | cut -d - -f 3-)
@@ -29,6 +33,10 @@ APT_SOURCES=/target/etc/apt/sources.list
 APT_CONFDIR=/target/etc/apt/apt.conf.d
 IT_CONFDIR=/target/etc/initramfs-tools/conf.d
 
+IFS_ORIG="$IFS"
+NL="
+"
+
 log() {
 	logger -t base-installer "$@"
 }
@@ -79,13 +87,15 @@ update_progress () {
 check_target () {
 	# Make sure something is mounted on the target.
 	# Partconf causes the latter format.
-	if ! grep -q '/target ' /proc/mounts && \
+	if [ -e /proc/mounts ] && \
+	   ! grep -q '/target ' /proc/mounts && \
 	   ! grep -q '/target/ ' /proc/mounts; then
 		exit_error base-installer/no_target_mounted
 	fi
 
 	# Warn about installation over an existing unix.
 	if [ -e /target/bin/sh -o -L /target/bin/sh ]; then
+		warning "attempting to install to unclean target"
 		db_capb ""
 		db_input high base-installer/use_unclean_target || true
 		db_go || exit 10
@@ -110,7 +120,7 @@ check_target () {
 	fi
 }
 
-setup_dev () {
+setup_dev_linux () {
 	# Ensure static device nodes created during install are preserved
 	# Tests in MAKEDEV require this is done in the D-I environment
 	mkdir -p /dev/.static/dev
@@ -120,40 +130,27 @@ setup_dev () {
 	mount --bind /dev /target/dev/
 }
 
-# TODO: as we no longer have to create devices here, the apt-install calls
-# could possibly better be done as post-base-installer hooks from partman
-install_filesystems () {
-	# RAID
-	if [ -e /proc/mdstat ] && grep -q ^md /proc/mdstat ; then
-		apt-install mdadm
-	fi
-	# device-mapper
-	if grep -q " device-mapper$" /proc/misc; then
-		# Avoid warnings from lvm2 tools about open file descriptors
-		export LVM_SUPPRESS_FD_WARNINGS=1
+setup_dev_kfreebsd() {
+	mount -t devfs devfs /target/dev
+}
 
-		# We can't check the root node directly as is done above because
-		# root could be on an LVM LV on top of an encrypted device
-		if type dmsetup >/dev/null 2>&1 && \
-		   dmsetup table | cut -d' ' -f4 | grep -q "crypt" 2>/dev/null; then
-			apt-install cryptsetup
-		fi
-
-		if type dmraid >/dev/null 2>&1; then
-			if [ "$(dmraid -s -c | grep -v "No RAID disks")" ]; then
-				apt-install dmraid
-			fi
-		fi
-
-		if pvdisplay | grep -iq "physical volume ---"; then
-			apt-install lvm2
-		fi
-	fi
+setup_dev() {
+	case "$OS" in
+		linux) setup_dev_linux ;;
+		kfreebsd) setup_dev_kfreebsd ;;
+		hurd) : ;;
+		*) warning "setup_dev called for an unknown OS ($OS)." ;;
+	esac	
 }
 
 configure_apt_preferences () {
 	[ ! -d "$APT_CONFDIR" ] && mkdir -p "$APT_CONFDIR"
 
+	# Don't install Recommends during base-installer
+	cat >$APT_CONFDIR/00InstallRecommends <<EOT
+APT::Install-Recommends "false";
+EOT
+
 	# Make apt trust Debian CDs. This is not on by default (we think).
 	# This will be left in place on the installed system.
 	cat > $APT_CONFDIR/00trustcdrom <<EOT
@@ -175,6 +172,14 @@ EOT
 	fi
 }
 
+final_apt_preferences () {
+	# From here on install Recommends as configured
+	db_get base-installer/install-recommends
+	if [ "$RET" = true ]; then
+		rm -f $APT_CONFDIR/00InstallRecommends
+	fi
+}
+
 apt_update () {
 	log-output -t base-installer chroot /target apt-get update \
 		|| apt_update_failed=$?
@@ -185,19 +190,23 @@ apt_update () {
 }
 
 install_extra () {
+	local IFS
 	info "Installing queued packages into /target/."
 
-	if [ -f /var/lib/apt-install/queue ] ; then
+	if [ -f /var/lib/apt-install/queue ]; then
 		# We need to install these one by one in case one fails.
-		PKG_COUNT=`cat /var/lib/apt-install/queue | wc -w`
+		PKG_COUNT=$(cat /var/lib/apt-install/queue | wc -w)
 		CURR_PKG=0
-		for PKG in `cat /var/lib/apt-install/queue`; do
+		IFS="$NL"
+		for LINE in $(cat /var/lib/apt-install/queue); do
+			IFS="$IFS_ORIG"
+			PKG=${LINE%% *}
+			OPTS=$(echo "$LINE" | sed "s/$PKG *//")
 			db_subst base-installer/section/install_extra_package SUBST0 "$PKG"
 			db_progress INFO base-installer/section/install_extra_package
 
-			if ! log-output -t base-installer apt-install $PKG; then
+			log-output -t base-installer apt-install $OPTS $PKG || \
 				warning "Failed to install $PKG into /target/: $?"
-			fi
 
 			# Advance progress bar within space allocated for install_extra
 			CURR_PKG=$(($CURR_PKG + 1))
@@ -316,7 +325,7 @@ get_mirror_info () {
 
 kernel_update_list () {
 	# Use 'uniq' to avoid listing the same kernel more then once
-	chroot /target apt-cache search '^(kernel|linux)-image' | \
+	chroot /target apt-cache search '^(kernel|linux|kfreebsd|gnumach)-image' | \
 	cut -d" " -f1 | uniq > "$KERNEL_LIST.unfiltered"
 	kernels=`sort -r "$KERNEL_LIST.unfiltered" | tr '\n' ' ' | sed -e 's/ $//'`
 	for candidate in $kernels; do
@@ -340,6 +349,8 @@ kernel_present () {
 
 pick_kernel () {
 	kernel_update_list
+	
+	db_settitle debian-installer/bootstrap-base/title
 
 	# For now, only present kernels we believe to be usable. We may have
 	# to rethink this later, but if there are no usable kernels it
@@ -443,7 +454,7 @@ pick_kernel () {
 	info "Using kernel '$KERNEL'"
 }
 
-install_linux () {
+install_kernel_linux () {
 	if [ "$KERNEL" = none ]; then
 		info "Not installing any kernel"
 		return
@@ -497,9 +508,7 @@ install_linux () {
 # Kernel image management overrides
 # See kernel-img.conf(5) for details
 do_symlinks = yes
-relative_links = yes
 do_bootloader = no
-do_bootfloppy = no
 do_initrd = $do_initrd
 link_in_boot = $link_in_boot
 EOF
@@ -559,11 +568,12 @@ EOF
 		# Select and set driver inclusion policy for initramfs-tools
 		if [ "$rd_generator" = initramfs-tools ]; then
 			if db_get base-installer/initramfs-tools/driver-policy && \
-			   [ "$RET" = "" ]; then
+			   [ -z "$RET" ]; then
 				# Get default for architecture
 				db_get base-installer/kernel/linux/initramfs-tools/driver-policy
 				db_set base-installer/initramfs-tools/driver-policy "$RET"
 			fi
+			db_settitle debian-installer/bootstrap-base/title
 			db_input medium base-installer/initramfs-tools/driver-policy || true
 			if ! db_go; then
 				db_progress stop
@@ -616,6 +626,12 @@ EOF
 		else
 			resume=
 		fi
+		if [ "$resume" ] && ! echo "$resume" | grep -q "^/dev/mapper/"; then
+			resume_uuid="$(block-attr --uuid "$resume" || true)"
+			if [ "$resume_uuid" ]; then
+				resume="UUID=$resume_uuid"
+			fi
+		fi
 		if [ -n "$resumeconf" ] && [ "$resume" ]; then
 			if [ -f $resumeconf ] ; then
 				sed -e "s@^#* *RESUME=.*@RESUME=$resume@" < $resumeconf > $resumeconf.new &&
@@ -640,7 +656,7 @@ EOF
 		fi
 	fi
 
-	# Advance progress bar to 30% of allocated space for install_linux
+	# Advance progress bar to 30% of allocated space for install_kernel_linux
 	update_progress 30 100
 
 	# Install the kernel
@@ -648,7 +664,7 @@ EOF
 	db_progress INFO base-installer/section/install_kernel_package
 	log-output -t base-installer apt-install "$KERNEL" || kernel_install_failed=$?
 
-	# Advance progress bar to 90% of allocated space for install_linux
+	# Advance progress bar to 90% of allocated space for install_kernel_linux
 	update_progress 90 100
 
 	if [ -f /target/etc/kernel-img.conf.$$ ]; then
@@ -669,6 +685,12 @@ get_resume_partition () {
 		if [ "$type" != partition ]; then
 			continue
 		fi
+		if [ ! -e "$filename" ]; then
+			continue
+		fi
+		if [ "${filename#/dev/ramzswap}" != "$filename" ]; then
+			continue
+		fi
 		if [ "$size" -gt "$biggest_size" ]; then
 			biggest_size="$size"
 			biggest_partition="$filename"
@@ -738,25 +760,136 @@ addmodule_yaird () {
 	fi
 }
 
+install_kernel_kfreebsd() {
+	if [ "$KERNEL" = none ]; then
+		info "Not installing any kernel"
+		return
+	fi
+
+	# Create configuration file for kernel-package
+	if [ -f /target/etc/kernel-img.conf ]; then
+		# Backup old kernel-img.conf
+		mv /target/etc/kernel-img.conf /target/etc/kernel-img.conf.$$
+	fi
+
+	cat > /target/etc/kernel-img.conf <<EOF
+# Kernel image management overrides
+# See kernel-img.conf(5) for details
+do_symlinks = no
+EOF
+	# Advance progress bar to 10% of allocated space for install_kfreebsd
+	update_progress 10 100
+
+	# Install the kernel
+	db_subst base-installer/section/install_kernel_package SUBST0 "$KERNEL"
+	db_progress INFO base-installer/section/install_kernel_package
+	log-output -t base-installer apt-install "$KERNEL" || kernel_install_failed=$?
+
+	# Advance progress bar to 90% of allocated space for install_kfreebsd
+	update_progress 90 100
+
+	if [ -f /target/etc/kernel-img.conf.$$ ]; then
+		# Revert old kernel-img.conf
+		mv /target/etc/kernel-img.conf.$$ /target/etc/kernel-img.conf
+	fi
+
+	if [ "$kernel_install_failed" ]; then
+		db_subst base-installer/kernel/failed-install KERNEL "$KERNEL"
+		exit_error base-installer/kernel/failed-install
+	fi
+}
+
+install_kernel_hurd() {
+	if [ "$KERNEL" = none ]; then
+		info "Not installing any kernel"
+		return
+	fi
+
+	# Create configuration file for kernel-package
+	if [ -f /target/etc/kernel-img.conf ]; then
+		# Backup old kernel-img.conf
+		mv /target/etc/kernel-img.conf /target/etc/kernel-img.conf.$$
+	fi
+
+	cat > /target/etc/kernel-img.conf <<EOF
+# Kernel image management overrides
+# See kernel-img.conf(5) for details
+do_symlinks = no
+EOF
+	# Advance progress bar to 10% of allocated space for install_kfreebsd
+	update_progress 10 100
+
+	# Install the kernel
+	db_subst base-installer/section/install_kernel_package SUBST0 "$KERNEL"
+	db_progress INFO base-installer/section/install_kernel_package
+	log-output -t base-installer apt-install "$KERNEL" || kernel_install_failed=$?
+
+	# Advance progress bar to 90% of allocated space for install_kernel_hurd
+	update_progress 90 100
+
+	if [ -f /target/etc/kernel-img.conf.$$ ]; then
+		# Revert old kernel-img.conf
+		mv /target/etc/kernel-img.conf.$$ /target/etc/kernel-img.conf
+	fi
+
+	if [ "$kernel_install_failed" ]; then
+		db_subst base-installer/kernel/failed-install KERNEL "$KERNEL"
+		exit_error base-installer/kernel/failed-install
+	fi
+}
+
+install_kernel() {
+	case "$OS" in
+		linux) install_kernel_linux ;;
+		kfreebsd) install_kernel_kfreebsd ;;
+		hurd) install_kernel_hurd ;;
+		*) warning "install_kernel called for an unknown OS ($OS)." ;;
+	esac	
+}
+
+
 # Assumes the file protocol is only used for CD (image) installs
 configure_apt () {
 	if [ "$PROTOCOL" = file ]; then
+		local tdir=/target/media$DIRECTORY
 		rm -f /var/lib/install-cd.id
 
 		# Let apt inside the chroot see the cdrom
-		if [ -n "$DIRECTORY" ]; then
-			umount /target$DIRECTORY 2>/dev/null || true
-			if [ ! -e /target/$DIRECTORY ]; then
-				mkdir -p /target/$DIRECTORY
-			fi
+		umount $tdir 2>/dev/null || true
+		if [ ! -e $tdir ]; then
+			mkdir -p $tdir
 		fi
 
 		# The bind mount is left mounted, for future apt-install
 		# calls to use.
-		if ! mount -o bind $DIRECTORY /target$DIRECTORY; then
-			warning "failed to bind mount /target$DIRECTORY"
-		fi
+		case "$OS" in
+			linux)
+			if ! mount -o bind $DIRECTORY $tdir; then
+				warning "failed to bind mount $tdir"
+			fi
+			;;
+			kfreebsd)
+			if ! mount -t nullfs $DIRECTORY $tdir ; then
+				warning "failed to bind mount $tdir"
+			fi
+			;;
+			hurd)
+			if ! mount -t firmlink $DIRECTORY $tdir > /dev/null 2>&1 ; then
+				warning "failed to bind mount $tdir"
+			fi
+			;;
+			*)
+			warning "configure_apt called with unknown OS ($OS)."
+			;;
+		esac
 
+		# Define the mount point for apt-cdrom
+		cat > $APT_CONFDIR/00CDMountPoint << EOT
+Acquire::cdrom {
+  mount "/media/cdrom";
+}
+Dir::Media::MountPath "/media/cdrom";
+EOT
 		# Make apt-cdrom and apt not unmount/mount CD-ROMs;
 		# needed to support CD images (hd-media installs).
 		# This file will be left in place until the end of the
@@ -765,11 +898,11 @@ configure_apt () {
 		cat > $APT_CONFDIR/00NoMountCDROM << EOT
 APT::CDROM::NoMount "true";
 Acquire::cdrom {
-  mount "/cdrom";
-  "/cdrom/" {
+  "/media/cdrom/" {
     Mount  "true";
     UMount "true";
   };
+  AutoDetect "false";
 }
 EOT
 
